/*******************************************************************************
*
* Description : Linux Device Driver to access PassMark PCIe loopback card
*
********************************************************************************/
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <asm/current.h>
#include <linux/pci_regs.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/kdev_t.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>

#include "pcie_common.h"

//Uncomment to print satus messages to DMESG
//#define DEBUG_MODE

#define DRV_NAME "PM_PCIE_DRIVER"
#define PCI_DEV_NAME "PM_PCIE_DEV"
#define DRV_VERSION	"2.1000"

#define PCI_BAR_0 0	
#define PCI_BAR_1 1

#define MAX_DMA_BUFFER 8192

// Maximum number of PCIE cards we will detect and communicate with
#define MAX_NUM_PCIECARDS 10 

//Register addresses: The /4 is due to the pointer arithmetic below (using 32bit integer pointers as that is what PCIE supports)

#define REG_IGL2_LED_CTRL 		0x8020	/ 4 	// LED register 
#define REG_IGL2_DMADIR			0x8308	/ 4 	// DMA_DIR register 
#define REG_IGL2_BASE_ADDR_WR		0x8208 	/ 4 	// Host -> Device
#define REG_IGL2_BASE_ADDR_RD_E1	0x8108 	/ 4 	// Device -> Host
#define REG_IGL2_BASE_ADDR_RD_E2	0x8110 	/ 4 	// Device -> Host
#define REG_IGL2_FW_VERSION		0x0008 	/ 4 	// Firmware version [31:4] Major [3:0] Minor
#define REG_IGL2_INT_ENABLED		0x8000 	/ 4	// Interrupts enabled/disabled (enabled by default)
#define REG_IGL2_XFER_SIZE		0x8300	/ 4	// Transfer size in 128B blocks

#define REG_IGL2_SER_NUM0		0x0010	/4 // Send when initialising driver. Write a '1' to retrieve serial number, then read this for lowest 32bits of serial numer
#define REG_IGL2_SER_NUM1		0x0018 	/ 4	// next 32bits of serial
#define REG_IGL2_SER_NUM2		0x0020	/ 4 	// next 32bits of serial
#define REG_IGL2_SER_NUM3		0x0028	/ 4	// last 32bits of serial
#define REG_IGL2_DMA_CH0_STATUS		0x8400 	/ 4	//DMA channel 0 status
#define REG_IGL2_DMA_CH1_STATUS		0x8408 	/ 4	//DMA channel 1 status
#define REG_IGL2_LSRAM_REG_DATA_LSB 	0x8048 / 4 	//LSRAM contents of index specified in LSRAM_REG_ADDRIOCTL_IGL2_ADC_STATS
#define REG_IGL2_LSRAM_REG_DATA_MSB 	0x009C / 4 	//LSRAM contents of index specified in LSRAM_REG_ADDR
#define REG_IGL2_LSRAM_REG_ADDR 	0x8040 / 4 	//LSRAM memory index address 

#define REG_IGL2_ADC_MODE		0x8500 / 4	//Change between ADC and stats mode  used for resetting values
#define REG_IGL2_ADC_CH_SEL		0x8510 / 4	//ADC channel selection register
#define REG_IGL2_ADC_INT_REG		0x8010 / 4	//Request ADC stat read?
#define	REG_IGL2_ADC_STAT_A0		0x8550 / 4	//Temperature sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_B5		0x8558 / 4	//3.3V aux sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_A2		0x8570 / 4	//12V pcie sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_B2		0x8578 / 4	//3.3V pcie sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_A3		0x8580 / 4	//12V pcie (Duplicate) sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_B3		0x8588 / 4	//12V molex sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_A4		0x8590 / 4	//5V molex sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_STAT_B4		0x8598 / 4	//12V molex (Duplicate) sensor min/max 27:16 Max, 11:0 Min
#define	REG_IGL2_ADC_AVG_0		0x85A0 / 4	//Average of 3.3Vaux (ch B5) 31:16, Avg of temp (ch A0) 15:0
#define	REG_IGL2_ADC_AVG_2		0x85A8 / 4	//Average of 3.3V (ch B2) 31:16, Avg of 12V (ch A2) 15:0
#define	REG_IGL2_ADC_AVG_3		0x85B0 / 4	//Average of 12V molex (ch B3) 31:16, Avg of 12V (ch A3) 15:0
#define	REG_IGL2_ADC_AVG_4		0x85B8 / 4	//Average of 12V molex (ch B4) 31:16, Avg of 5V molex (ch A4) 15:0
#define	REG_IGL2_LINK_WIDTH		0x0030 / 4	//Pcie link width, bit 0 = high is x4, low if x1
#define	REG_IGL2_LINK_MODE		0x0038 / 4	//Send when initialising driver.  Controls GEN2 LED bit 0 = high is link is GEN2, low if link is GEN1.

#define REG_IGL2_NUM_WRITE_ELEMENTS	0x8200	/ 4 	//Set to 2 for scatter gather, 1 for non-scatter gather
#define REG_IGL2_NUM_READ_ELEMENTS	0x8100	/ 4 	//Set to 2 for scatter gather, 1 for non-scatter gather


//LINK MODE Gen
#define LINK_MODE_GEN1	0x0
#define LINK_MODE_GEN2	0x1

//Stats mode
#define STATS_ON		0x11AA0001
#define STATS_OFF		0x11AA0000

//Register values
#define PCIE_CLEAR_DMA		0x11AA000F
#define PCIE_CLEAR_INT		0x11AA000F

//Using the same names as the windows driver to maintain consistancy
#define MODE_READ_ADC 		0x11AA0004 //ADC -> PCIe
#define MODE_READ_LSRAM		0x11AA0003 //LSRAM -> PCIe
#define MODE_WRITE_LSRAM 	0x11AA0008 //PCIe -> LSRAM
#define MODE_WRITE_LOOPBACK	0x11AA0009 //PCIe -> LSRAM -> PCIe

//ADC channel options
#define PCIE_ADC_12V_PCIE_BUF_AND_3V3_PCIE_BUF 	0
#define PCIE_ADC_12V_PCIE_BUF_AND_TEMP_SENS_BUF 1
#define PCIE_ADC_12V_PCIE_AND_3V3_PCIE		2
#define PCIE_ADC_12V_PCIE_AND_5V_MOLEX		3
#define PCIE_ADC_12V_MOLEX_AND_5V_MOLEX		4
#define PCIE_ADC_NUM_OPTIONS			5

#define SUCCESS 0
#define ENABLE 1
#define DISABLE 0



// channel status registers
typedef struct _DMA_CSR_
{
	unsigned int Count : 16 ; 	//Bit 0 -15
	unsigned int Reserved : 12 ; 	//Bit 16 - 27
	unsigned int IntFlag : 1 ; 	//Bit 28
	unsigned int Interrupt : 1 ; 	//Bit 29
	unsigned int Loopback : 1 ; 	//Bit 30 Only in DMA CSR ch 1
	unsigned int Done : 1; 		//Bit 31
}
DMA_CSR;

typedef struct _LINK_CAPS_
{
 	unsigned int MaxLinkSpeed : 4 ; 	//Bit 0 - 3
	unsigned int MaxLinkWidth : 6 ; 	//Bit 4 - 9
	unsigned int ASPM : 2 ; 	//Bit 10-11
	unsigned int L0ExitLatency : 3 ; 	//Bit 12-14
	unsigned int L1ExitLatency : 3 ; 	//Bit 15-17
	unsigned int LinkSpeedsVector : 7; 		//Bit 18-24
}
LINK_CAPS; 

typedef struct _PCIE_CAPS
{
 	unsigned int CapVersion : 4 ; 	//Bit 0 - 3
	unsigned int Type : 4; 	//Bit 4 - 7
	unsigned int Slot : 1 ; 	//Bit 8
	unsigned int IntMsgNumber : 5 ; 	//Bit 9-13
	unsigned int RSVD : 2 ; 	//Bit 14-15
}
PCIE_CAPS;

typedef struct _LINK_STATUS_
{
 	unsigned int CurLinkSpeed : 4 ; 	//Bit 0 - 3
	unsigned int CurLinkWidth : 6 ; 	//Bit 4 - 9
	unsigned int Undef : 1 ; 	//Bit 10
	unsigned int LinkTrain : 1 ; 	//Bit 11
	unsigned int SlotClock : 1; 	//Bit 12
	unsigned int DataLink : 1; 	//Bit 13
	unsigned int LBMS : 1; 	//Bit 14
	unsigned int LABS : 1; 	//Bit 15
}
LINK_STATUS ;


/* Structure to hold driver specific resources */
struct pm_pcie_dev {
	/* PCI BAR #0 */
	u32 bar0_base;
	u32 bar0_size;
	volatile void __iomem *base_addr0; 	// virtual address to represent PCI BAR #0

	/* PCI BAR #1 */
	u32 bar1_base;
	u32 bar1_size;
	volatile void __iomem *base_addr1; 	// virtual address to represent PCI BAR #1	 	

	dma_addr_t bus_addr_wr; 		// channel 0 HOST->DEVICE
	unsigned char* addr_wr;			// channel 0 HOST->DEVICE

	dma_addr_t bus_addr_rd;			// channel 1 DEVICE->HOST
	unsigned char* 	addr_rd;		// channel 1 DEVICE->HOST
	
	dev_t pci_dev_num; 			//Hold major/minor dev number
	struct pci_dev *pdev;
	struct cdev *pci_cdev;
 	struct device* device;

	unsigned int wr_buf[32]; 
	unsigned int rd_buf[32]; 
	unsigned int wr_buf_tmp[32];

	CONFIG_INFO currentConfig; 		//Save the current configuration information to be used during interrupts, read, writes etc
	wait_queue_head_t waitQueue; 
	bool interruptFinished ;
};



static unsigned long intCounter;
static unsigned int numPCICards = 0;
//Save the major device number recevied when registering device
static unsigned int pm_major_id = 0;

static struct class *pm_pcie_class = NULL;

	
struct dev_ext *pdevExt; 
static struct pm_pcie_dev * pm_pcie_cards = NULL; 

static struct pci_device_id ms_pci_tbl[] = {
	{ 0x11AA, 0x11AA, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{0,}
}; 

MODULE_DEVICE_TABLE(pci, ms_pci_tbl);


/**
* ms_pci_interrupt() - PCIe interrupt Handler
*
*  @irq : PCIe interrupt number
*/
static irqreturn_t ms_pci_interrupt(int irq, void* dev_id, struct pt_regs * regs)
{
	#ifdef DEBUG_MODE
	pr_info("%s: Got interrupt\n", DRV_NAME);
	#endif

	struct pm_pcie_dev* pm_pcie_card = (struct pm_pcie_dev*) dev_id;

	{
			
		{
			unsigned int* ptr1 = (unsigned int *)(pm_pcie_card->base_addr1);

			#ifdef DEBUG_MODE

			pr_info("%s: Interrupt intCounter = %ld \n", DRV_NAME, intCounter);

			//Read LSRAM temp debug
			pm_pcie_card->wr_buf_tmp[0] = 0; //Set read Index to 0wr_buf
			iowrite32(pm_pcie_card->wr_buf_tmp[0], (ptr1 + REG_IGL2_LSRAM_REG_ADDR)); 
			//Read LSRAM
			unsigned intLS1 = ioread32((ptr1 + REG_IGL2_LSRAM_REG_DATA_LSB));
			unsigned intLS2 = ioread32((ptr1 + REG_IGL2_LSRAM_REG_DATA_MSB));

			pr_info("%s: Interrupt LSRAM %x %x\n", DRV_NAME, intLS1, intLS2 );

			#endif

			//Clear the interrupt
			pm_pcie_card->wr_buf_tmp[0] = PCIE_CLEAR_INT;
			iowrite32(pm_pcie_card->wr_buf_tmp[0], (ptr1 + REG_IGL2_DMADIR)); 

			intCounter = intCounter + 1; /* To count interrupts */

			//wake up the process where it stopped (eg from a read/write)
			pm_pcie_card->interruptFinished = 1;
			wake_up_interruptible(&pm_pcie_card->waitQueue);

			#ifdef DEBUG_MODE
			pr_info("%s: returned IRQ_HANDLED\n", DRV_NAME);
			#endif

			return IRQ_HANDLED;
		}

	}

			#ifdef DEBUG_MODE
			pr_info("%s: returned IRQ_NONE\n", DRV_NAME);
			#endif
		return IRQ_NONE;
}   



/**
* ms_pci_remove() - Uninitialize the PCIe device
*
*  @pdev : Platform Device
*/
static void ms_pci_remove(struct pci_dev *pdev)
{
	//Gets called once for each card
	//Just do the device specific cleanup here 
	//memory cleanup happpens in driver unload function

	printk(KERN_INFO "%s: ms_pci_remove\n", DRV_NAME);
}


   
/**
* char_dev_open() - Open Char Device
*
*  @inode : Device Driver Information
*  @file  : Pointer to the file struct
*/
static int char_dev_open(struct inode *inode,
			    struct file  *file)
{
	unsigned int major = imajor(inode);
	unsigned int minor = iminor(inode);	
	struct pm_pcie_dev * dev = NULL;

	printk(KERN_INFO "%s: open operation invoked \n", DRV_NAME);
	//Need to get the major and minor numbers associated with this file and match one of the PCIE cads on the system

	if(major != pm_major_id || minor < 0 || minor >= numPCICards)
	{
		printk(KERN_INFO "%s: Open: No device found with major %d minor %d\n", DRV_NAME, major, minor);
		return -ENODEV;
	}

	//Store a pointer for other methods to use
	dev = &pm_pcie_cards[minor];
	file->private_data = dev;

	if(inode->i_cdev != dev->pci_cdev)
	{
		printk(KERN_INFO "%s: Open: Internal error, cdev mismatch\n", DRV_NAME);
		return -ENODEV;
	}

	
	return SUCCESS;
}

/**
* char_dev_release() - Release Char Device
*
*  @inode : Device Driver Information
*  @file  : Pointer to the file struct
*/

static int char_dev_release(struct inode *inode,
		            struct file *file)
{
	return SUCCESS;
}

static void clear_DMA_status(unsigned int *ptr1)
{
	//Clear the DMA status registers
	#ifdef DEBUG_MODE
	pr_info("%s: Clear DMA status\n", DRV_NAME);
	#endif

	iowrite32(PCIE_CLEAR_DMA, (ptr1 + REG_IGL2_DMADIR)); 

	#ifdef DEBUG_MODE
	pr_info("%s: Clear DMA status: wrote to %p\n", DRV_NAME, (ptr1 + REG_IGL2_DMADIR));
	#endif
}


/**
* char_dev_read() - Copy from PCIe memory to device file /dev/PM_PCI_DEV for user read
*
*  @file : Pointer to the file struct
*  @buf  : read buffer
*  @lbuf : size of the read buffer
*  @ppos : Position of buffer

Tim - can likely remove alot of these pr_info's below in productuion driver as 
they are pumping out a lot of debug info
*/

static ssize_t char_dev_read(struct file *file, 
                                     char *buf,
                                     size_t lbuf,
                                     loff_t *ppos)
{

	#ifdef DEBUG_MODE
	pr_info("%s: char_dev_read\n", DRV_NAME); 
	#endif

	//Need to get pointer to particular device from inode info
	struct pm_pcie_dev* res = (struct pm_pcie_dev*) file->private_data;
	int* ptr1 = NULL;
	int nbytes = 0; 

	if(buf == NULL)
	{
		pr_info("%s: NULL buffer passed\n", DRV_NAME);
		return -1;
	}

	if(lbuf < res->currentConfig.xferBufferSize)
	{
		pr_info("%s: User read buffer too small (%zu v %d)\n", DRV_NAME, lbuf, res->currentConfig.xferBufferSize);
		return -1;
	}

	#ifdef DEBUG_MODE
	pr_info("%s: char_dev_read: res->base_addr1 %p : user buf %p\n", DRV_NAME, res->base_addr1, buf); 
	#endif

	//In loopback mode just copy the DMA memory back to the user, should have already been updated as one instructio\n in the previous write
	//otherwise do similair to write function and clear / init dma direction / wait for interrupt
   	ptr1 = (unsigned int *)(res->base_addr1);

	if(res->currentConfig.writeMode != MODE_WRITE_LOOPBACK)
	{
		res->interruptFinished = 0;

	
	#ifdef DEBUG_MODE
	pr_info("%s: char_dev_read set REG_IGL2_DMADIR to %x\n", DRV_NAME, res->currentConfig.readMode); 
	#endif

		//Set the DMA direction to initiate the transfer		
		iowrite32(res->currentConfig.readMode, (ptr1 + REG_IGL2_DMADIR)); 
	
		//Need to wait for DMA transfer to suceed before returning
		wait_event_interruptible(res->waitQueue, res->interruptFinished != 0);

	}

	//Returns number of bytes not sucessfully copied, on success this should return 0
	nbytes = copy_to_user(buf, res->addr_rd, res->currentConfig.xferBufferSize);

	if( nbytes != 0 )
		pr_info("\n%s: ERROR nbytes %d\n", DRV_NAME, nbytes);

	return (res->currentConfig.xferBufferSize - nbytes);
}

/**
* char_dev_write() - Copy buffer from user and write to PCIe driver
*
*  @file : Pointer to the file struct
*  @buf  : Write buffer
*  @lbuf : size of the buffer write
*  @ppos : Position of buffer
*/


/**
 * 
 * @param inode 
 * @param filp 
 * @param cmd 
 * @param arg 
 * @return 
 */

//static int dev_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
static long dev_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	//Need to get pointer to particular device from inode info
	struct pm_pcie_dev* res = (struct pm_pcie_dev*) filp->private_data;
	
	int nbytes = 0;
	unsigned int *ptr0;
	unsigned int *ptr1;
	
	ptr0 = (unsigned int *)(res->base_addr0);
	ptr1 = (unsigned int *)(res->base_addr1);

	#ifdef DEBUG_MODE
	pr_info("\n%s: ptr0 = %p\n", DRV_NAME, ptr0); 
	pr_info("\n%s: ptr1 = %p\n", DRV_NAME, ptr1); 
	#endif
    
	//Need to cast long pointer to custom struct if we want more than 32 bits of argument
	
	switch (cmd)
	{

		case IOCTL_IGL2_GETCB:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_GETCB\n", DRV_NAME);	
			#endif

			//Returns number of bytes not sucessfully copied, on success this should return 0
			nbytes = copy_to_user(&arg, res->addr_rd, res->currentConfig.xferBufferSize);
			pr_info("\n%s: read %d of %d\n", DRV_NAME, res->currentConfig.xferBufferSize - nbytes, res->currentConfig.xferBufferSize);
		}
		break;


		case IOCTL_IGL2_LEDCTL: /* LED Access */
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_LEDCTL\n", DRV_NAME);
			#endif

			iowrite32(arg, (ptr1 + REG_IGL2_LED_CTRL)); 

		
		}
		break;

		case IOCTL_IGL2_CLEAR_DMA_REG:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_CLEAR_DMA_REG\n", DRV_NAME);
			#endif
			//Clear the DMA status registers
			clear_DMA_status(ptr1);
		}
		break;

		case IOCTL_IGL2_RESET_STATS_REG:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_RESET_STATS_REG\n", DRV_NAME);
			#endif

			iowrite32(STATS_OFF, (ptr1 + REG_IGL2_ADC_MODE));  //Disable stats mode
			wmb();
			#ifdef DEBUG_MODE
			int tmp = ioread32((ptr1 + REG_IGL2_ADC_MODE));
			pr_info("%s: REG_IGL2_ADC_MODE: 0x%x\n", DRV_NAME, tmp);
			#endif
			
			iowrite32(STATS_ON, (ptr1 + REG_IGL2_ADC_MODE));  //Restore to stats mode
			wmb();
			#ifdef DEBUG_MODE
			tmp = ioread32((ptr1 + REG_IGL2_ADC_MODE));
			pr_info("%s: REG_IGL2_ADC_MODE: 0x%x\n", DRV_NAME, tmp);
			#endif
		}
		break;

		case IOCTL_IGL2_SET_XFER_SIZE:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: REG_IGL2_XFER_SIZE (%d)\n", DRV_NAME, arg);
			#endif
			//Quick sanity chcek of transfer size
			if(arg <= 64)
			{
 				iowrite32(arg, (ptr1 + REG_IGL2_XFER_SIZE)); 
				//Keep track of current Xfersize
				res->currentConfig.xferSize = arg;
				res->currentConfig.xferBufferSize = arg * 128;
				
			}
		}
		break;
	
		case IOCTL_IGL2_GET_CONFIG:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_GET_CONFIG\n", DRV_NAME);
			#endif

			if(copy_to_user((CONFIG_INFO*) arg, &res->currentConfig, sizeof(CONFIG_INFO)) != 0)
			{
				pr_info("%s: Error IOCTL_IGL2_GET_CONFIG copy_to_user failed\n", DRV_NAME);	
				return -1;		
			}
		}
		break;

		case IOCTL_IGL2_SET_READ_MODE:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_SET_READ_MODE from %d to %d\n", DRV_NAME, res->currentConfig.readMode, arg);
			#endif

			// Change current read mode 
			// Actual DMA transfer is initiated later (during a read)
			switch(arg)
			{
				case PCIE_READ_ADC_MODE:
					res->currentConfig.readMode  = MODE_READ_ADC;
					break;
			
				case PCIE_READ_LSRAM_MODE:
					res->currentConfig.readMode  = MODE_READ_LSRAM;
					break;
			}

		}
		break;

	
		case IOCTL_IGL2_SET_WRITEMODE:
		{
			#ifdef DEBUG_MODE
			pr_info("%s: IOCTL_IGL2_SET_WRITEMODE from %d to %d\n", DRV_NAME, res->currentConfig.writeMode, arg);
			#endif

			// Change current read mode 
			// Actual DMA transfer is initiated later (during a write)
			switch(arg)
			{
				case PCIE_WRITE_BENCHMARK_MODE:
					res->currentConfig.writeMode = MODE_WRITE_LSRAM;			
					break;
			
				case PCIE_WRITE_LOOPBACK_MODE:
					res->currentConfig.writeMode = MODE_WRITE_LOOPBACK;
					break;
			}
		}
	break;

	
	case IOCTL_IGL2_SET_ADC_CH:
	{
		#ifdef DEBUG_MODE
		pr_info("%s: IOCTL_IGL2_SET_ADC_CH\n", DRV_NAME);
		#endif

		//Set the ADC channel
		#ifdef DEBUG_MODE
		pr_info("%s: Set ADC channel\n", DRV_NAME);
		#endif

		iowrite32(arg, (ptr1 + REG_IGL2_ADC_CH_SEL)); 

		#ifdef DEBUG_MODE
		pr_info("%s: wrote 0x%x to %p\n", DRV_NAME, arg, (ptr1+REG_IGL2_ADC_CH_SEL)); 
		#endif
			
			
	}
	break;

	//Read the stats value registers and write the output back to the file
	case IOCTL_IGL2_GET_ADC_STATS:
	{
		int tmp = 0;
		#ifdef DEBUG_MODE
		pr_info("%s: IOCTL_IGL2_GET_ADC_STATS\n", DRV_NAME);
		#endif

		ADC_STATS_RAW stats;
		memset(&stats, 0, sizeof(stats));

		//Request ADC stats read?
		tmp = ioread32(ptr1 + REG_IGL2_ADC_INT_REG);
		#ifdef DEBUG_MODE
		pr_info("%s: REG_IGL2_ADC_INT_REG 0x%x\n", DRV_NAME, tmp);

		//temp debug check ADC mode
		tmp = ioread32(ptr1 + REG_IGL2_ADC_MODE);
		pr_info("%s: REG_IGL2_ADC_MODE 0x%x\n", DRV_NAME, tmp);
		#endif

		stats.Stat_therm = ioread32(ptr1 + REG_IGL2_ADC_STAT_A0);
		stats.Stat_3v3aux = ioread32(ptr1 + REG_IGL2_ADC_STAT_B5);
		stats.Stat_12v = ioread32(ptr1 + REG_IGL2_ADC_STAT_A2);
		stats.Stat_3v3 = ioread32(ptr1 + REG_IGL2_ADC_STAT_B2);
		stats.Stat_12vpsu = ioread32(ptr1 + REG_IGL2_ADC_STAT_B3);
		stats.Stat_5vpsu = ioread32(ptr1 + REG_IGL2_ADC_STAT_A4);

		stats.Avg_1 = ioread32(ptr1 + REG_IGL2_ADC_AVG_0);
		stats.Avg_2 = ioread32(ptr1 + REG_IGL2_ADC_AVG_2);
		stats.Avg_3 = ioread32(ptr1 + REG_IGL2_ADC_AVG_3);
		stats.Avg_4 = ioread32(ptr1 + REG_IGL2_ADC_AVG_4);

		#ifdef DEBUG_MODE
		pr_info("%s: IOCTL_IGL2_GET_ADC_STATS\n", DRV_NAME);
		pr_info("%s: Stat_therm   %d\n", DRV_NAME, stats.Stat_therm);
		pr_info("%s: Stat_3v3aux  %d\n", DRV_NAME, stats.Stat_3v3aux);
		pr_info("%s: Stat_12v     %d\n", DRV_NAME, stats.Stat_12v);
		pr_info("%s: Stat_3v3     %d\n", DRV_NAME, stats.Stat_3v3);
		pr_info("%s: Stat_12vpsu  %d\n", DRV_NAME, stats.Stat_12vpsu);
		pr_info("%s: Stat_5vpsu   %d\n", DRV_NAME, stats.Stat_5vpsu);
		pr_info("%s: stats.Avg_1  %d\n", DRV_NAME, stats.Avg_1);
		pr_info("%s: stats.Avg_2  %d\n", DRV_NAME, stats.Avg_2);
		pr_info("%s: stats.Avg_3  %d\n", DRV_NAME, stats.Avg_3);
		pr_info("%s: stats.Avg_4  %d\n", DRV_NAME, stats.Avg_4);
		#endif

		if(copy_to_user((unsigned int*) arg, &stats, sizeof(stats)) != 0)
		{
			pr_info("%s: Error IOCTL_IGL2_GET_ADC_STATS copy_to_user failed\n", DRV_NAME);	
			return -1;		
		}
		
	}
	break;
	
	case IOCTL_IGL2_GET_THROUGHPUT:
	{
		#ifdef DEBUG_MODE
		pr_info("%s: IOCTL_IGL2_GET_THROUGHPUT\n", DRV_NAME);
		#endif

		THROUGHPUT_DATA throughputBuf;
		memset(&throughputBuf, 0, sizeof(throughputBuf));

		throughputBuf.Ch0Thruput = ioread32((ptr1 + REG_IGL2_DMA_CH0_STATUS));	
		throughputBuf.Ch1Thruput = ioread32((ptr1 + REG_IGL2_DMA_CH1_STATUS));	

		if(copy_to_user((THROUGHPUT_DATA*) arg, &throughputBuf, sizeof(throughputBuf)) != 0)
		{
			pr_info("%s: Error IOCTL_IGL2_GET_THROUGHPUT copy_to_user failed\n", DRV_NAME);	
			return -1;		
		}
		
	}
	break;

	default:
	{
		//unknown IOCTL
		pr_info("%s: Unknown IOCTL command: %x\n", DRV_NAME,cmd);
		return ENOTTY;
	}
	break;
	}
	return SUCCESS;
}

//Process a write command on the open device file, eg copy 
//contents to DMA buffer ready to transfer
static ssize_t char_dev_write(struct file *file, 
                                     const char *buf,
                                     size_t lbuf,
                                     loff_t *ppos)
{
	int* ptr1 = NULL;
	int nbytes = 0; 
	#ifdef DEBUG_MODE
	pr_info("%s: char_dev_write\n", DRV_NAME); 
	#endif

	//Need to get pointer to particular device from inode info
	struct pm_pcie_dev* res = (struct pm_pcie_dev*) file->private_data;


	if(buf == NULL)
	{
		#ifdef DEBUG_MODE
		pr_info("%s: NULL buffer passed\n", DRV_NAME);
		#endif

		return -1;
	}

	nbytes = copy_from_user(res->addr_wr,  buf, lbuf ); 
    
	#ifdef DEBUG_MODE
	int cnt = 0;
	for ( cnt = 0; cnt < 10; cnt++)
		pr_info("0x%x", res->addr_wr[cnt]);
	#endif

	res->interruptFinished = 0;

	//Clear the DMA status
   	 ptr1 = (unsigned int *)(res->base_addr1);
	//V2 commented out
	//clear_DMA_status(ptr1);

	//Set the DMA direction to initiate the transfer
	#ifdef DEBUG_MODE
	pr_info("%s: Start DMA transfer (%x) from write\n", DRV_NAME, res->currentConfig.writeMode); 
	#endif

	iowrite32(res->currentConfig.writeMode, (ptr1 + REG_IGL2_DMADIR)); 

	#ifdef DEBUG_MODE
	pr_info("%s: wait for interrupt from write\n", DRV_NAME, res->currentConfig.writeMode); 
	#endif
	//Need to wait for DMA transfer to suceed before returning
	wait_event_interruptible(res->waitQueue, res->interruptFinished != 0);

	return (res->currentConfig.xferBufferSize - nbytes);
	
}


static struct file_operations char_dev_fops = {
	.owner = THIS_MODULE,
	.open = char_dev_open,
	.write = char_dev_write,
	.read = char_dev_read,
	.release = char_dev_release,
	.unlocked_ioctl = dev_ioctl // This replaced .ioctl in kernels > 2.6.36
};

   
/**
* ms_pci_probe() - Initialize char device Module
*
*  @pdev : Platform Device
*  @ent  : PCI Device ID List Entry 
*  Resource mapping for PCIe Device
*/

//For each PCIe card detected/claimed by driver need to setup the resources / DMA areas etc
//And save details uin a global struct (already allocated)
static int ms_pci_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	int i = 0;
	unsigned int *ptr1 = NULL;
	unsigned int curVal = 0;
	int ret = 0;
	unsigned int serialnum[4];
	union 
	{
		LINK_STATUS bits;
		unsigned short  ushort;
	} LinkStatus;
	bool finished = false;
	unsigned char capPointer = 0;
	unsigned char capType = 0;
	int result = 0;
	union 
	{
		PCIE_CAPS bits;
		unsigned short  ushort;
	} PCIeCaps;
	
	union 
	{
		LINK_CAPS bits;
		unsigned int uint ;
	} LinkCaps;
	int temp32Val = 0;
	unsigned short tmp16Val = 0;


	memset(serialnum, 0, sizeof(serialnum));
	//Check we havne't already claimed the max number of cards
	if(numPCICards > MAX_NUM_PCIECARDS)
	{
	   	pr_info("%s: max number of PCIe cards already claimed by driver\n", DRV_NAME);
		return -ENODEV;
	}


	//Initialise wait queue
	init_waitqueue_head(&pm_pcie_cards[numPCICards].waitQueue);

        pr_info("%s: pci device %d found %p\n", DRV_NAME, numPCICards, pdev);

 	i = pci_enable_device(pdev);
	if (i)
		return i;

	pci_set_master(pdev);


	//Setup device id major/minor info
	pm_pcie_cards[numPCICards].pci_dev_num = MKDEV(pm_major_id, numPCICards);
	
   	  pr_info("%s: pci_dev_num %d\n", DRV_NAME, pm_pcie_cards[numPCICards].pci_dev_num);


	// PCI BAR 0# resources 
	pm_pcie_cards[numPCICards].bar0_base = pci_resource_start (pdev, PCI_BAR_0);
	pr_info("%s: PCI BAR #0 BASE = 0x%x\n", DRV_NAME, pm_pcie_cards[numPCICards].bar0_base);

	if (!pm_pcie_cards[numPCICards].bar0_base || ((pci_resource_flags (pdev, PCI_BAR_0) & IORESOURCE_MEM) == 0)) {
                dev_err(&pdev->dev, "%s: no I/O resource at PCI BAR #0\n", DRV_NAME);
        }
    else {
	   pm_pcie_cards[numPCICards].bar0_size = pci_resource_len (pdev, PCI_BAR_0);
       pr_info("%s: PCI BAR #0 SIZE = 0x%x\n", DRV_NAME, pm_pcie_cards[numPCICards].bar0_size);
    }

	if(request_mem_region(pm_pcie_cards[numPCICards].bar0_base, pm_pcie_cards[numPCICards].bar0_size, DRV_NAME) == NULL) {
		pr_info("%s: PCI BAR #0 Space not available\n", DRV_NAME);
	}
    else {
         pm_pcie_cards[numPCICards].base_addr0 = (void *)ioremap_nocache(pm_pcie_cards[numPCICards].bar0_base, pm_pcie_cards[numPCICards].bar0_size); 
         pr_info("%s: base_addr0 = 0x%p\n", DRV_NAME, pm_pcie_cards[numPCICards].base_addr0);
    }
	if(pm_pcie_cards[numPCICards].base_addr0 == NULL) {
		pr_info("%s: Unable to map BAR #0 address", DRV_NAME);
	}
		
	 // PCI BAR 1# resources 

        pm_pcie_cards[numPCICards].bar1_base = pci_resource_start (pdev, PCI_BAR_1);
        pr_info("%s: PCI BAR #1 BASE = 0x%x\n", DRV_NAME, pm_pcie_cards[numPCICards].bar1_base);

        if (!pm_pcie_cards[numPCICards].bar1_base || ((pci_resource_flags (pdev, PCI_BAR_1) & IORESOURCE_MEM) == 0)) {
                dev_err(&pdev->dev, "%s: no I/O resource at PCI BAR #1\n", DRV_NAME);
                goto err_out_free_res;
        }

        pm_pcie_cards[numPCICards].bar1_size = pci_resource_len (pdev, PCI_BAR_1);
        pr_info("%s: PCI BAR #1 SIZE = 0x%x\n", DRV_NAME, pm_pcie_cards[numPCICards].bar1_size);

        if(request_mem_region(pm_pcie_cards[numPCICards].bar1_base, pm_pcie_cards[numPCICards].bar1_size, DRV_NAME) == NULL) {
                pr_info("%s: PCI BAR #1 Space not available\n", DRV_NAME);
                goto err_out_free_res;
        }

/*
        pm_pcie_cards[numPCICards].base_addr1 = (void *)ioremap_nocache(pm_pcie_cards[numPCICards].bar1_base, pm_pcie_cards[numPCICards].bar1_size);
        pr_info("%s: base_addr1 = 0x%p\n", DRV_NAME, pm_pcie_cards[numPCICards].base_addr1);
        if(pm_pcie_cards[numPCICards].base_addr1 == NULL) {
                pr_info("%s: Unable to map BAR #1 address", DRV_NAME);
                goto err_out_free_reg;
        }
*/
        pm_pcie_cards[numPCICards].base_addr1 = (void *)pci_iomap(pdev,1,  pm_pcie_cards[numPCICards].bar1_size);
        pr_info("%s: base_addr1 (pci_iomap) = 0x%p\n", DRV_NAME, pm_pcie_cards[numPCICards].base_addr1);
        if(pm_pcie_cards[numPCICards].base_addr1 == NULL) {
                pr_info("%s: Unable to map BAR #1 address", DRV_NAME);
                goto err_out_free_reg;
        }


    	ptr1 = (unsigned int*)pm_pcie_cards[numPCICards].base_addr1;


	//Allocate DMA buffers

	//make sure we only allocate 32bit addresses
	ret = pci_set_dma_mask(pdev, DMA_BIT_MASK(32));
	pr_info("%s: pci_set_dma_mask returned %d\n", DRV_NAME, ret);

	ret = pci_set_consistent_dma_mask(pdev, DMA_BIT_MASK(32));
	pr_info("%s: pci_set_consistent_dma_mask returned %d\n", DRV_NAME, ret);
	

	//pm_pcie_cards[numPCICards].addr_wr = pci_alloc_consistent(pdev, MAX_DMA_BUFFER, &pm_pcie_cards[numPCICards].bus_addr_wr); 
	//pm_pcie_cards[numPCICards].addr_rd = pci_alloc_consistent(pdev, MAX_DMA_BUFFER, &pm_pcie_cards[numPCICards].bus_addr_rd); 
	pm_pcie_cards[numPCICards].addr_wr = dma_alloc_coherent(NULL, MAX_DMA_BUFFER, &pm_pcie_cards[numPCICards].bus_addr_wr, GFP_KERNEL); 
	pm_pcie_cards[numPCICards].addr_rd = dma_alloc_coherent(NULL, MAX_DMA_BUFFER, &pm_pcie_cards[numPCICards].bus_addr_rd, GFP_KERNEL); 

	if(pm_pcie_cards[numPCICards].addr_wr == NULL || pm_pcie_cards[numPCICards].addr_rd == NULL)
	{
		pr_info("%s: ERROR unable to allocate memory\n", DRV_NAME);
		 goto err_out_free_res;
	}

	//pr_info("%s: Allocated DMA memory addr_wr %p %p : addr_rd %p %p\n", DRV_NAME, 
	//	pm_pcie_cards[numPCICards].addr_wr, pm_pcie_cards[numPCICards].bus_addr_wr, pm_pcie_cards[numPCICards].addr_rd, pm_pcie_cards[numPCICards].bus_addr_rd);

	//write DMA addresses to specific registers
	//curVal = ioread32((ptr1 + REG_IGL2_BASE_ADDR_RD));
	//pr_info("%s: REG_IGL2_BASE_ADDR_RD %p old value %p\n", DRV_NAME, (ptr1 + REG_IGL2_BASE_ADDR_RD), curVal);

	//pr_info("%s: REG_IGL2_BASE_ADDR_RD set %p to DMA addr_rd = 0x%p (0x%p)\n", DRV_NAME, (ptr1 + REG_IGL2_BASE_ADDR_RD), pm_pcie_cards[numPCICards].bus_addr_rd, curVal);
	curVal = pm_pcie_cards[numPCICards].bus_addr_rd;
	iowrite32(curVal, (ptr1+ REG_IGL2_BASE_ADDR_RD_E1));  

	//curVal = ioread32((ptr1 + REG_IGL2_BASE_ADDR_RD));
	//pr_info("%s: REG_IGL2_BASE_ADDR_RD %p current value %p\n", DRV_NAME, (ptr1 + REG_IGL2_BASE_ADDR_RD), curVal);

	//curVal = ioread32((ptr1 + REG_IGL2_BASE_ADDR_WR));
	//pr_info("%s: REG_IGL2_BASE_ADDR_WR %p old value %p\n", DRV_NAME, (ptr1 + REG_IGL2_BASE_ADDR_WR), curVal);
	//pr_info("%s: REG_IGL2_BASE_ADDR_WR set %p to DMA addr_rd = 0x%p (0x%p)\n", DRV_NAME, (ptr1 + REG_IGL2_BASE_ADDR_WR), pm_pcie_cards[numPCICards].bus_addr_wr, curVal);

	curVal = pm_pcie_cards[numPCICards].bus_addr_wr;
	iowrite32(curVal, (ptr1+ REG_IGL2_BASE_ADDR_WR));  

	//curVal = ioread32((ptr1 + REG_IGL2_BASE_ADDR_WR));
	//pr_info("%s: REG_IGL2_BASE_ADDR_WR %p current value %p\n", DRV_NAME, (ptr1 + REG_IGL2_BASE_ADDR_WR), curVal);

	//Set transfer size to 32
	curVal = 32;
	iowrite32(curVal, (ptr1+ REG_IGL2_XFER_SIZE));  

	//Get serial number and print to log
	//Send signal to get serial number
	curVal = 1;
	iowrite32(curVal, (ptr1+ REG_IGL2_SER_NUM0));  

	msleep(300);

	serialnum[0] = ioread32((ptr1 + REG_IGL2_SER_NUM0));
	serialnum[1] = ioread32((ptr1 + REG_IGL2_SER_NUM1));
	serialnum[2] = ioread32((ptr1 + REG_IGL2_SER_NUM2));
	serialnum[3] = ioread32((ptr1 + REG_IGL2_SER_NUM3));
	

	pr_info("%s: Serial#: 0x%x %x %x %x\n",DRV_NAME, serialnum[0], serialnum[1], serialnum[2], serialnum[3]);
	pm_pcie_cards[numPCICards].currentConfig.sn0 = serialnum[0];
	pm_pcie_cards[numPCICards].currentConfig.sn1 = serialnum[1];
	pm_pcie_cards[numPCICards].currentConfig.sn2 = serialnum[2];
	pm_pcie_cards[numPCICards].currentConfig.sn3 = serialnum[3];	

	//Get firmware version
	result = ioread32((ptr1+ REG_IGL2_FW_VERSION));
	pr_info("%s: Firmware: 0x%x\n",DRV_NAME, result);	
	pm_pcie_cards[numPCICards].currentConfig.fwVersion = result;

	//Get link width
	result = ioread32((ptr1+ REG_IGL2_LINK_WIDTH));
	pr_info("%s: Link Width REG: 0x%x\n",DRV_NAME, result);	
	pm_pcie_cards[numPCICards].currentConfig.linkWidth = result;
	

	//Enable MSI interrupts
	if(pci_enable_msi(pdev))
	{
		pr_info("%s: Unable to enable MSI interrupt 0x%x\n",DRV_NAME, pdev->irq);
		goto err_out_free_reg;	
	}

	pr_info("%s: MSI interrupt 0x%x\n", DRV_NAME, pdev->irq);

	if(request_irq(pdev->irq, (irq_handler_t)ms_pci_interrupt, IRQF_SHARED, DRV_NAME, (void*)&pm_pcie_cards[numPCICards]) ) 
	{
		pr_info("%s: unable to register irq 0x%x\n", DRV_NAME, pdev->irq);
        	goto err_out_free_irq;
	} 
	

	pr_info("%s: Create /dev file\n", DRV_NAME);
	//create device file

 if (!( pm_pcie_cards[numPCICards].pci_cdev = cdev_alloc ())) 
	{
        	  printk (KERN_ERR "%s: cdev_alloc() failed\n", DRV_NAME);
         //   unregister_chrdev_region (ms_pci_dev, MAX_NUM_PCIECARDS);
            	return -1;
 	}
	
	pr_info("%s: cdev_init\n", DRV_NAME);
	cdev_init( pm_pcie_cards[numPCICards].pci_cdev, &char_dev_fops);

	pr_info("%s: cdev_add\n", DRV_NAME);
	ret=cdev_add( pm_pcie_cards[numPCICards].pci_cdev, pm_pcie_cards[numPCICards].pci_dev_num, MAX_NUM_PCIECARDS);
	if( ret < 0 ) 
	{
		printk(KERN_INFO "%s: Error registering device driver\n", DRV_NAME);
	        cdev_del ( pm_pcie_cards[numPCICards].pci_cdev);
                //unregister_chrdev_region (ms_pci_dev, MAX_NUM_PCIECARDS); 	
		return -1;
	}

	pr_info("%s: device_create\n", DRV_NAME);
	pm_pcie_cards[numPCICards].device = device_create (pm_pcie_class, NULL, pm_pcie_cards[numPCICards].pci_dev_num, NULL, PCI_DEV_NAME "%d", numPCICards);

	if(IS_ERR(pm_pcie_cards[numPCICards].device ))
	{
		int err = PTR_ERR(pm_pcie_cards[numPCICards].device);
		pr_info("%s: Error %d while trying to create device %s%d \n", DRV_NAME, err, PCI_DEV_NAME, numPCICards);
		return -1;
	}
	
	printk("%s: Successfully created /dev/%s%d\n", DRV_NAME, PCI_DEV_NAME, numPCICards);






	
	result = pci_read_config_byte(pdev, 0x34, &capPointer); 

	//Read PCIE configuration and determine what vesion of PCIe is in use (1, 2, 3) 
	//Currently using the link status register
	while(finished == false)
	{
		result = pci_read_config_byte(pdev, capPointer, &capType); 

		if(result != 0)
			break;

		pr_info("%s: Struct type %x\n", DRV_NAME, capType);
		if(capType == 0x10)
		{
			 finished = true;
		}			
		else
		{
			result = pci_read_config_byte(pdev, capPointer+1, &capPointer); 
		pr_info("%s: NExt offset %x\n", DRV_NAME, capPointer);
		}
			
	} 


	if(finished == true)
	{
	result = pci_read_config_word(pdev, capPointer+0x12, &LinkStatus.ushort);
	pr_info("%s: Link status register 0x%X - Link speed: %d Link width: %d\n", DRV_NAME, LinkStatus.ushort, LinkStatus.bits.CurLinkSpeed, LinkStatus.bits.CurLinkWidth);
	//Save pcie version
	pm_pcie_cards[numPCICards].currentConfig.pcieVer = LinkStatus.bits.CurLinkSpeed;
	}
	else
	{
		pm_pcie_cards[numPCICards].currentConfig.pcieVer = 0;
	}

	result = pci_read_config_dword(pdev, capPointer+0xC, &LinkCaps.uint);
	//pr_info("%s: PCIE capabilites: 0x%X : Version 0x%X\n", DRV_NAME, PCIeCaps.ushort, PCIeCaps.bits.CapVersion);
	pr_info("%s: PCIE capabilites pointer: 0x%X\n", DRV_NAME, capPointer);


	result = pci_read_config_word(pdev, capPointer+0x2, &PCIeCaps.ushort);
	pr_info("%s: PCIe cap register 0x%X : version %d\n", DRV_NAME, PCIeCaps.ushort, PCIeCaps.bits.CapVersion);

	result = pci_read_config_dword(pdev, capPointer+0x4, &temp32Val);
	pr_info("%s: Dev cap register 0x%X\n", DRV_NAME, temp32Val);

	//If dev cap register bits 0:3  == 1, use LinkCaps.bits.MaxLinkSpeed
	//if > 1 then try and read Link caps 2, ifthat is 0 (as seen on old linux dev)
	// then use bits 0:3 of Root Complex Link capabilites and check for ==1 (2.5Gt/s) or ==2 (5.0 Gt/s & 2.5Gts)

	result = pci_read_config_word(pdev, capPointer+0x8, &tmp16Val);
	pr_info("%s: Dev control register 0x%X\n", DRV_NAME, tmp16Val);

	result = pci_read_config_word(pdev, capPointer+0xA, &tmp16Val);
	pr_info("%s: Dev status register 0x%X\n", DRV_NAME, tmp16Val);


	result = pci_read_config_word(pdev, capPointer+0x12, &tmp16Val);
	pr_info("%s: Link status register 0x%X\n", DRV_NAME, tmp16Val);


	pr_info("%s: Link capabilites 0x%X : Max Link Speed 0x%X\n", DRV_NAME, LinkCaps.uint, LinkCaps.bits.MaxLinkSpeed);
	
	result = pci_read_config_dword(pdev, capPointer+0x2C, &LinkCaps.uint);
	pr_info("%s: Link capabilites 2 0x%X : \n", DRV_NAME, LinkCaps.uint);
	

	//Check for pcie v2+ and set LINK_MODE register
	if( LinkStatus.bits.CurLinkSpeed > 1)
	{
		pr_info("%s: Set link mode GEN2\n", DRV_NAME);
		iowrite32(LINK_MODE_GEN2, (ptr1+ REG_IGL2_LINK_MODE));
	}
	else
	{
		pr_info("%s: Set link mode GEN1\n", DRV_NAME);	
		iowrite32(LINK_MODE_GEN1, (ptr1+ REG_IGL2_LINK_MODE));
	}

	//Enable interrupts
	pr_info("%s: Enable interrupts \n", DRV_NAME);
	curVal = 0x7;
	iowrite32(curVal, (ptr1+ REG_IGL2_INT_ENABLED));  

	pm_pcie_cards[numPCICards].pdev = pdev;

	pr_info("%s: IOCTLS \n", DRV_NAME);
	pr_info("%s: IOCTL_IGL2_LEDCTL %lx \n", DRV_NAME, IOCTL_IGL2_LEDCTL);
	pr_info("%s: IOCTL_IGL2_GET_THROUGHPUT %lx \n", DRV_NAME, IOCTL_IGL2_GET_THROUGHPUT);
	pr_info("%s: IOCTL_IGL2_SET_READ_MODE %lx \n", DRV_NAME, IOCTL_IGL2_SET_READ_MODE);
	pr_info("%s: IOCTL_IGL2_SET_ADC_CH %lx \n", DRV_NAME, IOCTL_IGL2_SET_ADC_CH);
	pr_info("%s: IOCTL_IGL2_SET_WRITEMODE %lx \n", DRV_NAME, IOCTL_IGL2_SET_WRITEMODE);
	pr_info("%s: IOCTL_IGL2_CLEAR_DMA_REG %x \n", DRV_NAME, IOCTL_IGL2_CLEAR_DMA_REG);
	pr_info("%s: IOCTL_IGL2_GETCB %lx \n", DRV_NAME, IOCTL_IGL2_GETCB);
	pr_info("%s: IOCTL_IGL2_SET_XFER_SIZE %lx \n", DRV_NAME, IOCTL_IGL2_SET_XFER_SIZE);
	pr_info("%s: IOCTL_IGL2_GET_CONFIG %lx \n", DRV_NAME, IOCTL_IGL2_GET_CONFIG);
	pr_info("%s: IOCTL_IGL2_READ_LSRAM_REG %lx \n", DRV_NAME, IOCTL_IGL2_READ_LSRAM_REG);
	pr_info("%s: IOCTL_IGL2_READ_REG %lx \n", DRV_NAME, IOCTL_IGL2_READ_REG);
	pr_info("%s: IOCTL_IGL2_ACCESS_CONFIG_SPACE %x \n", DRV_NAME, IOCTL_IGL2_ACCESS_CONFIG_SPACE);
	pr_info("%s: IOCTL_IGL2_GET_ADC_STATS %lx \n", DRV_NAME, IOCTL_IGL2_GET_ADC_STATS);
	pr_info("%s: IOCTL_IGL2_SET_VOLT_TOL %lx \n", DRV_NAME, IOCTL_IGL2_SET_VOLT_TOL);
	pr_info("%s: IOCTL_IGL2_RESET_STATS_REG %x \n", DRV_NAME, IOCTL_IGL2_RESET_STATS_REG);

	
	pm_pcie_cards[numPCICards].currentConfig.xferSize = ioread32(ptr1 + REG_IGL2_XFER_SIZE);
	pm_pcie_cards[numPCICards].currentConfig.xferBufferSize = pm_pcie_cards[numPCICards].currentConfig.xferSize * 128;
	pm_pcie_cards[numPCICards].currentConfig.ledStatus  = ioread32((ptr1 + REG_IGL2_LED_CTRL));;


	//Increment num cards found 
	numPCICards++;

	return 0;

err_out_free_irq:
	free_irq(pdev->irq, &pm_pcie_cards[numPCICards]);
	//pci_disable_msi(pdev);
	iounmap(pm_pcie_cards[numPCICards].base_addr0);
	iounmap(pm_pcie_cards[numPCICards].base_addr1);
err_out_free_reg:
	release_mem_region(pm_pcie_cards[numPCICards].bar0_base, pm_pcie_cards[numPCICards].bar0_size);
	release_mem_region(pm_pcie_cards[numPCICards].bar1_base, pm_pcie_cards[numPCICards].bar1_size);
err_out_free_res:
	//kfree((const void *)res);	
	return -ENODEV;
}


static struct pci_driver ms_pci_driver = {
        .name           = DRV_NAME,
        .probe          = ms_pci_probe,
        .remove         = ms_pci_remove,
        .id_table       = ms_pci_tbl,
};

/**
*  pcidrv_init() - Initialize PCIe Module
*
*  Registers Platform Driver
*/

static __init int pcidrv_init(void)
{
	dev_t ms_pci_dev;


	pr_info("%s: Version %s Module init \n", DRV_NAME, DRV_VERSION);
	
	if (alloc_chrdev_region (&ms_pci_dev, 0, MAX_NUM_PCIECARDS, PCI_DEV_NAME) < 0) {
            printk (KERN_ERR "%s: failed to reserve major/minor range\n", DRV_NAME);
            return -1;
    	}

	printk("%s: Got Major %d\n", DRV_NAME, MAJOR(ms_pci_dev) );

	//Save major ID received
	pm_major_id = MAJOR(ms_pci_dev) ;

        pm_pcie_class = class_create (THIS_MODULE, "VIRTUAL");
        
	//Allocate memory for pcei card structure
	 pm_pcie_cards = (struct pm_pcie_dev*) kzalloc(MAX_NUM_PCIECARDS * sizeof(struct pm_pcie_dev), GFP_KERNEL);
	if(pm_pcie_cards == NULL)
	{
		return -ENOMEM;
	}

	if(pci_register_driver(&ms_pci_driver))
	{
		pr_info("%s: unable to register pci driver \n", DRV_NAME);
		return -1;
	}

	return 0;
}

/**
* pcidrv_cleanup() - Uninitialize the PCIe Module
*
*  Unregister the Platform Driver
*/

static __exit void  pcidrv_cleanup(void)
{	
	//Need to update to account for multiple cards????
	int count = 0;
	int curVal = 0;
	dev_t ms_pci_dev = MKDEV(pm_major_id, 0);
	unsigned int* ptr1 = NULL;
	
	pr_info("%s: Module being removed \n", DRV_NAME);
	
	//Disable  interrupts on all cards

	for( count = 0; count < numPCICards && count < MAX_NUM_PCIECARDS; count++)
	{
		pr_info("%s: %d Stop interrupts \n", DRV_NAME, count);
		ptr1 = (unsigned int *)(pm_pcie_cards[count].base_addr1);
		iowrite32(curVal, (ptr1 + REG_IGL2_INT_ENABLED));  		

		free_irq(pm_pcie_cards[count].pdev->irq, (void*)&pm_pcie_cards[count]);

		pr_info("%s: %d Disable device \n", DRV_NAME, count);
		pci_disable_msi(pm_pcie_cards[count].pdev);
 		pci_clear_master(pm_pcie_cards[count].pdev);
		pci_disable_device(pm_pcie_cards[count].pdev);

		//clean up device memory now they have been closed
		pr_info("%s: %d Clean up memory \n", DRV_NAME, count);
		iounmap(pm_pcie_cards[count].base_addr0);
		iounmap(pm_pcie_cards[count].base_addr1);
		release_mem_region(pm_pcie_cards[count].bar0_base, pm_pcie_cards[count].bar0_size);
		release_mem_region(pm_pcie_cards[count].bar1_base, pm_pcie_cards[count].bar1_size);

		dma_free_coherent(NULL, MAX_DMA_BUFFER, pm_pcie_cards[count].addr_wr, pm_pcie_cards[count].bus_addr_wr);
		dma_free_coherent(NULL, MAX_DMA_BUFFER, pm_pcie_cards[count].addr_rd, pm_pcie_cards[count].bus_addr_rd);

		//pci_free_consistent(pm_pcie_cards[count].pdev, MAX_DMA_BUFFER, pm_pcie_cards[count].addr_wr, pm_pcie_cards[count].bus_addr_wr);
		//pci_free_consistent(pm_pcie_cards[count].pdev, MAX_DMA_BUFFER, pm_pcie_cards[count].addr_rd, pm_pcie_cards[count].bus_addr_rd);

 		cdev_del( pm_pcie_cards[count].pci_cdev);

	 	device_destroy (pm_pcie_class, pm_pcie_cards[count].pci_dev_num);
	}

	kfree((const void *)pm_pcie_cards);

	//Unregister the driver, forces the remove call to all devices 
	pr_info("%s: Unregister driver \n", DRV_NAME);
	pci_unregister_driver (&ms_pci_driver);

	pr_info("%s: Class cleanup \n", DRV_NAME);

        class_destroy (pm_pcie_class);
	unregister_chrdev_region(ms_pci_dev, MAX_NUM_PCIECARDS);
	printk(KERN_INFO "%s: Driver unregistered \n", DRV_NAME);
}

module_init(pcidrv_init);
module_exit(pcidrv_cleanup);

MODULE_AUTHOR("PassMark");
MODULE_DESCRIPTION("PCI device driver");
MODULE_LICENSE("GPL");
